# strptime.rb: Written by Tadayoshi Funaba 1999, 2000
# $Id: strptime.rb,v 1.8 2000-09-24 19:14:53+09 tadf Exp $

class Date; end

class << Date

  MONTHS = {
    'january'  => 1, 'february' => 2, 'march'    => 3, 'april'    => 4,
    'may'      => 5, 'june'     => 6, 'july'     => 7, 'august'   => 8,
    'september'=> 9, 'october'  =>10, 'november' =>11, 'december' =>12,
    'jan'      => 1, 'feb'      => 2, 'mar'      => 3, 'apr'      => 4,
    'may'      => 5, 'jun'      => 6, 'jul'      => 7, 'aug'      => 8,
    'sep'      => 9, 'oct'      =>10, 'nov'      =>11, 'dec'      =>12
  }
  MONTHPAT = MONTHS.keys.join('|')

  DAYS = {
    'sunday'   => 0, 'monday'   => 1, 'tuesday'  => 2, 'wednesday'=> 3,
    'thursday' => 4, 'friday'   => 5, 'saturday' => 6,
    'sun'      => 0, 'mon'      => 1, 'tue'      => 2, 'wed'      => 3,
    'thu'      => 4, 'fri'      => 5, 'sat'      => 6
  }
  DAYPAT = DAYS.keys.join('|')

  ZONES = {
    'ut' =>  0*3600, 'gmt'=>  0*3600, 'est'=> -5*3600, 'edt'=> -4*3600,
    'cst'=> -6*3600, 'cdt'=> -5*3600, 'mst'=> -7*3600, 'mdt'=> -6*3600,
    'pst'=> -8*3600, 'pdt'=> -7*3600,
    'a'  => -1*3600, 'b'  => -2*3600, 'c'  => -3*3600, 'd'  => -4*3600,
    'e'  => -5*3600, 'f'  => -6*3600, 'g'  => -7*3600, 'h'  => -8*3600,
    'i'  => -9*3600, 'k'  =>-10*3600, 'l'  =>-11*3600, 'm'  =>-12*3600,
    'n'  =>  1*3600, 'o'  =>  2*3600, 'p'  =>  3*3600, 'q'  =>  4*3600,
    'r'  =>  5*3600, 's'  =>  6*3600, 't'  =>  7*3600, 'u'  =>  8*3600,
    'v'  =>  9*3600, 'w'  => 10*3600, 'x'  => 11*3600, 'y'  => 12*3600,
    'z'  =>  0*3600,
    'uct'=>  0*3600, 'utc'=>  0*3600, 'jst'=>  9*3600
  }

  def strptime2(str, fmt, elem)
    fmt.scan(/%[EO]?.|./o) do |c|
      cc = c.sub(/\A%[EO]?(.)\Z/o, '%\\1')
      case cc
      when /\A\s/o
        str.sub!(/\A\s+/o, '')
      when '%A', '%a'
	return unless str.sub!(/\A(#{DAYPAT})\b/io, '')
	val = DAYS[$1.downcase]
	return unless val
	elem[:wday] = val
      when '%B', '%b'
	return unless str.sub!(/\A(#{MONTHPAT})\b/io, '')
 	val = MONTHS[$1.downcase]
	return unless val
 	elem[:mon] = val
      when '%C';
      when '%c'
	return unless strptime2(str, '%a %b %e %H:%M:%S %Y', elem)
      when '%D'
	return unless strptime2(str, '%m/%d/%y', elem)
      when '%d', '%e'
	return unless str.sub!(/\A(\d+)/o, '')
	val = $1.to_i
	return unless (1..31) === val
	elem[:mday] = val
      when '%F'
	return unless strptime2(str, '%Y-%m-%d', elem)
      when '%G'
	return unless str.sub!(/\A(\d+)/o, '')
	val = $1.to_i
	elem[:cwyear] = val
      when '%g'
	return unless str.sub!(/\A(\d+)/o, '')
	val = $1.to_i
        val += if val >= 69 then 1900 else 2000 end
	elem[:cwyear] = val
      when '%H'
	return unless str.sub!(/\A(\d+)/o, '')
	val = $1.to_i
	return unless (0..23) === val
	elem[:hour] = val
      when '%h'
	return unless strptime2(str, '%b', elem)
      when '%I'
	return unless str.sub!(/\A(\d+)/o, '')
	val = $1.to_i
	return unless (1..12) === val
	elem[:hour] = val % 12
      when '%j'
	return unless str.sub!(/\A(\d+)/o, '')
	val = $1.to_i
	return unless (1..366) === val
	elem[:yday] = val
      when '%k'
	return unless strptime2(str, '%H', elem)
      when '%l'
	return unless strptime2(str, '%I', elem)
      when '%M'
	return unless str.sub!(/\A(\d+)/o, '')
	val = $1.to_i
	return unless (0..59) === val
	elem[:min] = val
      when '%m'
	return unless str.sub!(/\A(\d+)/o, '')
	val = $1.to_i
	return unless (1..12) === val
	elem[:mon] = val
      when '%n'
	str.sub!(/\A /o, '')
      when '%P'
	return unless strptime2(str, '%p', elem)
      when '%p'
	return unless str.sub!(/\A([ap])\.?m\.?\b/io, '')
	elem[:pm] = $1.downcase == 'p'
      when '%R'
	return unless strptime2(str, '%H:%M', elem)
      when '%r'
	return unless strptime2(str, '%I:%M:%S %p', elem)
      when '%S'
	return unless str.sub!(/\A(\d+)/o, '')
	val = $1.to_i
	return unless (0..60) === val
	elem[:sec] = val
      when '%s'
	return unless str.sub!(/\A(\d+)/o, '')
	val = $1.to_i
	elem[:seconds] = val
      when '%T'
	return unless strptime2(str, '%H:%M:%S', elem)
      when '%t'
	str.sub!(/\A /o, '')
      when '%U', '%W'
	return unless str.sub!(/\A(\d+)/o, '')
	val = $1.to_i
	return unless (0..53) === val
	elem[if c == '%U' then :wnum0 else :wnum1 end] = val
      when '%u'
	return unless str.sub!(/\A(\d+)/o, '')
	val = $1.to_i
	return unless (1..7) === val
	elem[:cwday] = val
      when '%V'
	return unless str.sub!(/\A(\d+)/o, '')
	val = $1.to_i
	return unless (1..53) === val
	elem[:cweek] = val
      when '%v'
	return unless strptime2(str, '%e-%b-%Y', elem)
      when '%w'
	return unless str.sub!(/\A(\d+)/o, '')
	val = $1.to_i
	return unless (0..6) === val
	elem[:wday] = val
      when '%X'
	return unless strptime2(str, '%H:%M:%S', elem)
      when '%x'
	return unless strptime2(str, '%m/%d/%y', elem)
      when '%Y'
	return unless str.sub!(/\A(\d+)/o, '')
	val = $1.to_i
	elem[:year] = val
      when '%y'
	return unless str.sub!(/\A(\d+)/o, '')
	val = $1.to_i
        val += if val >= 69 then 1900 else 2000 end
	elem[:year] = val
      when '%Z'
	return unless str.sub!(/\A([a-z0-9:+-]+(?:\s+dst\b)?)/io, '')
	val = $1
	elem[:zone] = val
	if ZONES[val.downcase]
	  off = ZONES[val.downcase]
	elsif /\A([+-])(\d{2}):?(\d{2})\Z/no =~ val
	  sign = if $1 == '+' then +1 else -1 end
	  off = sign * ($2.to_i * 3600 + $3.to_i * 60)
	elsif /\A([+-])(\d{2})\Z/no =~ val
	  sign = if $1 == '+' then +1 else -1 end
	  off = sign * ($2.to_i * 3600)
	end
	elem[:off] = off
      when '%z'
	return unless strptime2(str, '%Z', elem)
      when '%%'
	return unless str.sub!(/\A%/o, '')
      when '%+'
	return unless strptime2(str, '%a %b %e %H:%M:%S %Z %Y', elem)
      else
	return unless str.sub!(Regexp.new('\\A' + Regexp.quote(c)), '')
      end
    end
    if elem.delete :pm
      if elem[:hour]
	elem[:hour] += 12
      end
    end
    str
  end

  private :strptime2

  def strptime(str, fmt='%c')
    elem = {}
    elem if strptime2(str.dup, fmt, elem)
  end

end
